#!/usr/bin/env python3
#
# fontconfig/doc/edit-sgml.py
#
# Copyright © 2003 Keith Packard
# Copyright © 2020 Tim-Philipp Müller
#
# Permission to use, copy, modify, distribute, and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice appear in all copies and that both that
# copyright notice and this permission notice appear in supporting
# documentation, and that the name of the author(s) not be used in
# advertising or publicity pertaining to distribution of the software without
# specific, written prior permission.  The authors make no
# representations about the suitability of this software for any purpose.  It
# is provided "as is" without express or implied warranty.
#
# THE AUTHOR(S) DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
# EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

import argparse
import sys
import re

parser = argparse.ArgumentParser()
parser.add_argument('template')
parser.add_argument('input')
parser.add_argument('output')

args = parser.parse_known_args()

template_fn = args[0].template
output_fn = args[0].output
input_fn = args[0].input

# -------------
# Read template
# -------------

with open(template_fn, 'r', encoding='utf8') as f:
  template_text = f.read()

template_lines = template_text.strip().split('\n')

# -------------------------------------
# Read replacement sets from .fncs file
# -------------------------------------

replacement_sets = []

# TODO: also allow '-' for stdin
with open(input_fn, 'r', encoding='utf8') as f:
  fncs_text = f.read()

# split into replacement sets
fncs_chunks = fncs_text.strip().split('@@')

for chunk in fncs_chunks:
  # get rid of any preamble such as license and FcFreeTypeQueryAll decl in fcfreetype.fncs
  start = chunk.find('@')
  if start:
    chunk = chunk[start:]

  # split at '@' and remove empty lines (keep it simple instead of doing fancy
  # things with regular expression matches, we control the input after all)
  lines = [line for line in chunk.split('@') if line.strip()]

  replacement_set = {}

  while lines:
    tag = lines.pop(0).strip()
    # FIXME: this hard codes the tag used in funcs.sgml - we're lazy
    if tag.startswith('PROTOTYPE'):
      text = ''
    else:
      text = lines.pop(0).strip()
      if text.endswith('%'):
        text = text[:-1] + ' '

    replacement_set[tag] = text

  if replacement_set:
    replacement_sets += [replacement_set]

# ----------------
# Open output file
# ----------------

if output_fn == '-':
  fout = sys.stdout
else:
  fout = open(output_fn, "w", encoding='utf8')

# ----------------
# Process template
# ----------------

def do_replace(template_lines, rep, tag_suffix=''):
  skip_tag = None
  skip_lines = False
  loop_lines = []
  loop_tag = None

  for t_line in template_lines:
    # This makes processing easier and is the case for our templates
    if t_line.startswith('@') and not t_line.endswith('@'):
      sys.exit('Template lines starting with @ are expected to end with @, please fix me!')

    if loop_tag:
      loop_lines += [t_line]

    # Check if line starts with a directive
    if t_line.startswith('@?'):
      tag = t_line[2:-1] + tag_suffix
      if skip_tag:
        sys.exit('Recursive skipping not supported, please fix me!')
      skip_tag = tag
      skip_lines = tag not in rep
    elif t_line.startswith('@:'):
      if not skip_tag:
        sys.exit('Skip else but no active skip list?!')
      skip_lines = skip_tag in rep
    elif t_line.startswith('@;'):
      if not skip_tag:
        sys.exit('Skip end but no active skip list?!')
      skip_tag = None
      skip_lines = False
    elif t_line.startswith('@{'):
      if loop_tag or tag_suffix != '':
        sys.exit('Recursive looping not supported, please fix me!')
      loop_tag = t_line[2:-1]
    elif t_line.startswith('@}'):
      tag = t_line[2:-1] + tag_suffix
      if not loop_tag:
        sys.exit('Loop end but no active loop?!')
      if loop_tag != tag:
        sys.exit(f'Loop end but loop tag mismatch: {loop_tag} != {tag}!')
      loop_lines.pop() # remove loop end directive
      suffix = '+'
      while loop_tag + suffix in rep:
        do_replace(loop_lines, rep, suffix)
        suffix += '+'
      loop_tag = None
      loop_lines = []
    else:
      if not skip_lines:
        # special-case inline optional substitution (hard-codes specific pattern in funcs.sgml because we're lazy)
        output_line = re.sub(r'@\?(RET)@@RET@@:@(void)@;@', lambda m: rep.get(m.group(1) + tag_suffix, m.group(2)), t_line)
        # replace any substitution tags with their respective substitution text
        output_line = re.sub(r'@(\w+)@', lambda m: rep.get(m.group(1) + tag_suffix, ''), output_line)
        print(output_line, file=fout)

# process template for each replacement set
for rep in replacement_sets:
  do_replace(template_lines, rep)



#! /usr/bin/env python3
# Copyright (C) 2025 fontconfig Authors
# SPDX-License-Identifier: HPND

import argparse
import glob
import re
import sys
from pathlib import Path

parser = argparse.ArgumentParser()
parser.add_argument('srcdir')
parser.add_argument('builddir')

args = parser.parse_args()
h_apis = []
d_apis = []

header_files = glob.glob(str(Path(args.builddir) / 'fontconfig' / '*.h')) + glob.glob(str(Path(args.srcdir) / 'fontconfig' / '*.h'))

for fn in header_files:
    with open(fn) as f:
        for s in f:
            if re.search(r'^Fc', s) and not re.search('FcPublic', s):
                res = re.sub(r'[^a-zA-Z0-9].*', '', s)
                h_apis += [res]
            m = re.search(r'#\s*define\s+(Fc[a-zA-Z]+)\s*\(.*$', s)
            if m:
                h_apis += [m.group(1)]

h_apis.sort()

doc_files = glob.glob(str(Path(args.srcdir) / 'doc' / '*.fncs'))
for fn in doc_files:
    with open(fn) as f:
        for s in f:
            if re.search(r'@FUNC[+]*@', s):
                d_apis += [s.split()[1]]

d_apis.sort()

missing_in_doc = list(set(h_apis) - set(d_apis))
missing_in_header = list(set(d_apis) - set(h_apis))
if len(missing_in_doc) == 0 and len(missing_in_header) == 0:
    sys.exit(0)

if len(missing_in_doc) > 0:
    print('Missing API documentation:')
    for s in missing_in_doc:
        print(f'\t{s}')
    print('')
if len(missing_in_header) > 0:
    print('API available in documentation only:')
    for s in missing_in_header:
        print(f'\t{s}')
    print('')

sys.exit(1)



#!/usr/bin/env python3
#
# fontconfig/doc/extract-man-list.py
#
# Parses .fncs files and extracts list of man pages that will be generated
#
# Copyright © 2020 Tim-Philipp Müller
#
# Permission to use, copy, modify, distribute, and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice appear in all copies and that both that
# copyright notice and this permission notice appear in supporting
# documentation, and that the name of the author(s) not be used in
# advertising or publicity pertaining to distribution of the software without
# specific, written prior permission.  The authors make no
# representations about the suitability of this software for any purpose.  It
# is provided "as is" without express or implied warranty.
#
# THE AUTHOR(S) DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
# EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

import sys
import re

replacement_sets = []

# -------------------------------------
# Read replacement sets from .fncs file
# -------------------------------------

def read_fncs_file(fn):
  global replacement_sets

  with open(fn, 'r', encoding='utf8') as f:
    fncs_text = f.read()

    # split into replacement sets
    fncs_chunks = fncs_text.strip().split('@@')

    for chunk in fncs_chunks:
      # get rid of any preamble such as license and FcFreeTypeQueryAll decl in fcfreetype.fncs
      start = chunk.find('@')
      if start:
        chunk = chunk[start:]

      # split at '@' and remove empty lines (keep it simple instead of doing fancy
      # things with regular expression matches, we control the input after all)
      lines = [line for line in chunk.split('@') if line.strip()]

      replacement_set = {}

      while lines:
        tag = lines.pop(0).strip()
        # FIXME: this hard codes the tag used in funcs.sgml - we're lazy
        if tag.startswith('PROTOTYPE'):
          text = ''
        else:
          text = lines.pop(0).strip()
          if text.endswith('%'):
            text = text[:-1] + ' '

        replacement_set[tag] = text

      if replacement_set:
        replacement_sets += [replacement_set]

# ----------------------------------------------------------------------------
#  Main
# ----------------------------------------------------------------------------

if len(sys.argv) < 2:
  sys.exit('Usage: {} FILE1.FNCS [FILE2.FNCS...]'.format(sys.argv[0]))

fout = sys.stdout

for input_fn in sys.argv[1:]:
  read_fncs_file(input_fn)

# process template for each replacement set
for rep in replacement_sets:
  if 'FUNC+' in rep:
    man_page_title = rep.get('TITLE', rep['FUNC'])
  else:
    man_page_title = rep['FUNC']
  print(man_page_title)



#!/usr/bin/env python3
#
# fontconfig/fc-case/fc-case.py
#
# Copyright © 2004 Keith Packard
# Copyright © 2019 Tim-Philipp Müller
#
# Permission to use, copy, modify, distribute, and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice appear in all copies and that both that
# copyright notice and this permission notice appear in supporting
# documentation, and that the name of the author(s) not be used in
# advertising or publicity pertaining to distribution of the software without
# specific, written prior permission.  The authors make no
# representations about the suitability of this software for any purpose.  It
# is provided "as is" without express or implied warranty.
#
# THE AUTHOR(S) DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
# EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

from enum import Enum
import argparse
import string
import sys


class CaseFoldClass(Enum):
    COMMON = 1
    FULL = 2
    SIMPLE = 3
    TURKIC = 4


class CaseFoldMethod(Enum):
    RANGE = 0
    EVEN_ODD = 1
    FULL = 2


caseFoldClassMap = {
  'C': CaseFoldClass.COMMON,
  'F': CaseFoldClass.FULL,
  'S': CaseFoldClass.SIMPLE,
  'T': CaseFoldClass.TURKIC
}

folds = []

def ucs4_to_utf8(ucs4):
    utf8_rep = []

    if ucs4 < 0x80:
        utf8_rep.append(ucs4)
        bits = -6
    elif ucs4 < 0x800:
        utf8_rep.append(((ucs4 >> 6) & 0x1F) | 0xC0)
        bits = 0
    elif ucs4 < 0x10000:
        utf8_rep.append(((ucs4 >> 12) & 0x0F) | 0xE0)
        bits = 6
    elif ucs4 < 0x200000:
        utf8_rep.append(((ucs4 >> 18) & 0x07) | 0xF0)
        bits = 12
    elif ucs4 < 0x4000000:
        utf8_rep.append(((ucs4 >> 24) & 0x03) | 0xF8)
        bits = 18
    elif ucs4 < 0x80000000:
        utf8_rep.append(((ucs4 >> 30) & 0x01) | 0xFC)
        bits = 24
    else:
        return []

    while bits >= 0:
        utf8_rep.append(((ucs4 >> bits) & 0x3F) | 0x80)
        bits -= 6

    return utf8_rep


def utf8_size(ucs4):
    return len(ucs4_to_utf8(ucs4))


case_fold_method_name_map = {
    CaseFoldMethod.RANGE: 'FC_CASE_FOLD_RANGE,',
    CaseFoldMethod.EVEN_ODD: 'FC_CASE_FOLD_EVEN_ODD,',
    CaseFoldMethod.FULL: 'FC_CASE_FOLD_FULL,',
}

if __name__=='__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('case_folding_file')
    parser.add_argument('--template', dest='template_file', default=None)
    parser.add_argument('--output', dest='output_file', default=None)

    args = parser.parse_args()

    minFoldChar = None
    maxFoldChar = None
    fold = None

    foldChars = []
    maxFoldChars = 0

    maxExpand = 0

    # Read the standard Unicode CaseFolding.txt file
    with open(args.case_folding_file, 'r', encoding='utf-8') as casefile:
        for cnt, line in enumerate(casefile):
            if not line or not line[0] in string.hexdigits:
                continue

            # print('Line {}: {}'.format(cnt, line.strip()))

            tokens = line.split('; ')

            if len(tokens) < 3:
                print('Not enough tokens in line {}'.format(cnt),
                      file=sys.stderr)
                sys.exit(1)

            # Get upper case value
            upper = int(tokens.pop(0), 16)

            # Get class
            cfclass = caseFoldClassMap[tokens.pop(0)]

            # Get list of result characters
            lower = list(map(lambda s: int(s, 16), tokens.pop(0).split()))

            # print('\t----> {:04X} {} {}'.format(upper, cfclass, lower))

            if not minFoldChar:
                minFoldChar = upper

            maxFoldChar = upper

            if cfclass in [CaseFoldClass.COMMON, CaseFoldClass.FULL]:
                if len(lower) == 1:
                    # foldExtends
                    if fold and fold['method'] == CaseFoldMethod.RANGE:
                        foldExtends = (lower[0] - upper) == fold['offset'] and upper == fold['upper'] + fold['count']
                    elif fold and fold['method'] == CaseFoldMethod.EVEN_ODD:
                        foldExtends = (lower[0] - upper) == 1 and upper == (fold['upper'] + fold['count'] + 1)
                    else:
                        foldExtends = False

                    if foldExtends:
                        # This modifies the last fold item in the array too
                        fold['count'] = upper - fold['upper'] + 1
                    else:
                        fold = {}
                        fold['upper'] = upper
                        fold['offset'] = lower[0] - upper
                        if fold['offset'] == 1:
                            fold['method'] = CaseFoldMethod.EVEN_ODD
                        else:
                            fold['method'] = CaseFoldMethod.RANGE
                        fold['count'] = 1
                        folds.append(fold)
                    expand = utf8_size(lower[0]) - utf8_size(upper)
                else:
                    fold = {}
                    fold['upper'] = upper
                    fold['method'] = CaseFoldMethod.FULL
                    fold['offset'] = len(foldChars)

                    # add chars
                    for c in lower:
                        utf8_rep = ucs4_to_utf8(c)
                        # print('{} -> {}'.format(c,utf8_rep))
                        for utf8_char in utf8_rep:
                            foldChars.append(utf8_char)

                    fold['count'] = len(foldChars) - fold['offset']
                    folds.append(fold)

                    if fold['count'] > maxFoldChars:
                        maxFoldChars = fold['count']

                    expand = fold['count'] - utf8_size(upper)
                    if expand > maxExpand:
                        maxExpand = expand

    # Open output file
    if args.output_file:
        sys.stdout = open(args.output_file, 'w', encoding='utf-8')

    print('/*')
    print(' * This file was generated against CaseFolding.txt from'
          ' Unicode.org.')
    print(' * All the data in array is a part of them and licensed'
          ' under a terms of use:')
    print(' *   https://www.unicode.org/terms_of_use.html')
    print(' */')
    print('')
    # Dump these tables
    print('#define FC_NUM_CASE_FOLD\t{}'.format(len(folds)))
    print('#define FC_NUM_CASE_FOLD_CHARS\t{}'.format(len(foldChars)))
    print('#define FC_MAX_CASE_FOLD_CHARS\t{}'.format(maxFoldChars))
    print('#define FC_MAX_CASE_FOLD_EXPAND\t{}'.format(maxExpand))
    print('#define FC_MIN_FOLD_CHAR\t0x{:08x}'.format(minFoldChar))
    print('#define FC_MAX_FOLD_CHAR\t0x{:08x}'.format(maxFoldChar))
    print('')

    # Dump out ranges
    print('static const FcCaseFold    fcCaseFold[FC_NUM_CASE_FOLD] = {')
    for f in folds:
        short_offset = f['offset']
        if short_offset < -32367:
            short_offset += 65536
        if short_offset > 32368:
            short_offset -= 65536
        print(f'    {{ 0x{f["upper"]:08x}, '
              f'{case_fold_method_name_map[f["method"]]:22s} '
              f'0x{f["count"]:04x}, {short_offset:6d} }},')
    print('};\n')

    # Dump out "other" values
    print('static const FcChar8\tfcCaseFoldChars[FC_NUM_CASE_FOLD_CHARS] = {')
    for n, c in enumerate(foldChars):
        if n == len(foldChars) - 1:
            end = ''
        elif n % 16 == 15:
            end = ',\n'
        else:
            end = ','
        print('0x{:02x}'.format(c), end=end)
    print('\n};')

    sys.stdout.flush()



#!/usr/bin/env python3
#
# gperf extract_index_column.py
#
# Copyright (C) 2020 Tim-Philipp Müller <tim centricular com>
#
# This file is part of GNU GPERF.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Takes .gperf file and strips the 'struct' bit at the top and all but the
# first column of data, e.g. to extract just foreign_name from lang-utf8.gperf

import os
import sys

if len(sys.argv) < 2:
  print('ERROR: usage: {} INPUT.GPERF [OUTPUT.GPERF]'.format(sys.argv[0]))
  sys.exit(-1)

input_fn = sys.argv[1]
f = open(input_fn, 'r', encoding='utf8')

if len(sys.argv) > 2 and sys.argv[2] != '-':
  f_out = open(sys.argv[2], 'w', encoding='utf8')
else:
  f_out = sys.stdout

seen_hdr_sep = False

for (n, (line)) in enumerate(f):
  line = line.strip()
  if not seen_hdr_sep:
    seen_hdr_sep = line == '%%'
    continue
  first_column = line.split(',')[0]
  print(first_column, file=f_out)

sys.exit(0)



#! /usr/bin/env python3
# Copyright (C) 2024 fontconfig Authors
# SPDX-License-Identifier: HPND

import os
import pytest
import re
import requests
import shutil
import subprocess
from pathlib import Path


def test_issue431(tmp_path):
    builddir = Path(os.environ.get("builddir", Path(__file__).parent.parent))
    roboto_flex_font = (
        builddir
        / "testfonts"
        / "roboto-flex-fonts/fonts/variable/RobotoFlex[GRAD,XOPQ,XTRA,YOPQ,YTAS,YTDE,YTFI,YTLC,YTUC,opsz,slnt,wdth,wght].ttf"
    )

    if not roboto_flex_font.exists():
        pytest.skip(f"Font file not found: {roboto_flex_font}")

    result = subprocess.run(
        [
            builddir / "fc-query" / "fc-query",
            "-f",
            "%{family[0]}:%{index}:%{style[0]}:%{postscriptname}\n",
            roboto_flex_font,
        ],
        stdout=subprocess.PIPE,
    )

    for line in result.stdout.decode("utf-8").splitlines():
        family, index, style, psname = line.split(":")
        normstyle = re.sub("[\x04\\(\\)/<>\\[\\]{}\t\f\r\n ]", "", style)
        assert (
            psname.split("-")[-1] == normstyle
        ), f"postscriptname `{psname}' does not contain style name `{normstyle}': index {index}"



#!/usr/bin/env python3

from pathlib import Path
import argparse
import re

if __name__== '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('input')
    parser.add_argument('output')
    args = parser.parse_args()

    input_lines = Path(args.input).read_text(encoding='utf-8').splitlines()
    with Path(args.output).open('w', encoding='utf-8') as out:
        write = True
        for l in input_lines:
            if l.startswith('CUT_OUT_BEGIN'):
                write = False

            if write and l:
                stripped = re.sub(r'^\s+', '', l)
                stripped = re.sub(r'\s*,\s*', ',', stripped)
                if not stripped.isspace() and stripped:
                    out.write('%s\n' % stripped)

            if l.startswith('CUT_OUT_END'):
                write = True



#!/usr/bin/env python3

import os
import re
import sys
import argparse
from collections import OrderedDict

# cat fontconfig/fontconfig.h | grep '^Fc[^ ]* *(' | sed -e 's/ *(.*$//'

def extract(fname):
    with open(fname, 'r', encoding='utf-8') as f:
        for l in f.readlines():
            l = l.rstrip()
            m = re.match(r'^(Fc[^ ]*)[\s\w]*\(.*', l)

            if m and m.group(1) not in ['FcCacheDir', 'FcCacheSubdir']:
                yield m.group(1)

if __name__=='__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('srcdir')
    parser.add_argument('head')
    parser.add_argument('tail')
    parser.add_argument('headers', nargs='+')

    args = parser.parse_args()

    definitions = {}

    for fname in os.listdir(args.srcdir):
        define_name, ext = os.path.splitext(fname)
        if ext != '.c':
            continue

        define_name = '__%s__' % os.path.basename(define_name)

        for definition in extract(os.path.join(args.srcdir, fname)):
            definitions[definition] = define_name

    declarations = OrderedDict()

    for fname in args.headers:
        for declaration in extract(fname):
            try:
                define_name = definitions[declaration]
            except KeyError:
                print ('error: could not locate %s in src/*.c' % declaration)
                sys.exit(1)

            declarations[declaration] = define_name

    with open(args.head, 'w') as head:
        with open(args.tail, 'w') as tail:
            tail.write('#if HAVE_GNUC_ATTRIBUTE\n')
            last = None
            for name, define_name in declarations.items():
                alias = 'IA__%s' % name
                hattr = 'FC_ATTRIBUTE_VISIBILITY_HIDDEN'
                head.write('extern __typeof (%s) %s %s;\n' % (name, alias, hattr))
                head.write('#define %s %s\n' % (name, alias))
                if define_name != last:
                    if last is not None:
                        tail.write('#endif /* %s */\n' % last)
                    tail.write('#ifdef %s\n' % define_name)
                    last = define_name
                tail.write('# undef %s\n' % name)
                cattr = '__attribute((alias("%s"))) FC_ATTRIBUTE_VISIBILITY_EXPORT' % alias
                tail.write('extern __typeof (%s) %s %s;\n' % (name, name, cattr))
            tail.write('#endif /* %s */\n' % last)
            tail.write('#endif /* HAVE_GNUC_ATTRIBUTE */\n')



#! /usr/bin/env python3
# Copyright (C) 2025 Google LLC.
# SPDX-License-Identifier: HPND

import os
from pathlib import Path
import pytest
import re
import requests
import subprocess


def builddir():
    return Path(os.environ.get("builddir", Path(__file__).parent.parent))


def list_test_fonts():
    font_files = []
    for root, _, files in os.walk(builddir() / "testfonts"):
        for file in files:
            # Variable .ttc not supported yet.
            if file.endswith(".ttf"):
                font_files.append(os.path.join(root, file))
    return font_files


def run_fc_query(font_file, with_fontations=False):
    fc_query_path = builddir() / "fc-query" / "fc-query"

    env = os.environ.copy()
    if with_fontations:
        env["FC_FONTATIONS"] = ""

    result = subprocess.run(
        [fc_query_path, font_file],
        stdout=subprocess.PIPE,
        env=env,
        stderr=subprocess.PIPE,
        text=True,
        check=False,
    )

    assert (
        result.returncode == 0
    ), f"fc-query failed for {font_file} with error: {result.stderr}"
    assert result.stdout, f"fc-query produced no output for {font_file}"

    return result


@pytest.mark.parametrize("font_file", list_test_fonts())
def test_fontations_freetype_fcquery_equal(font_file):
    print(f"Testing with: {font_file}")  # Example usage

    font_path = Path(font_file)

    if not font_path.exists():
        pytest.skip(f"Font file not found: {font_file}")  # Skip if file missing

    result_freetype = run_fc_query(font_file).stdout.strip().splitlines()
    result_fontations = (
        run_fc_query(font_file, with_fontations=True)
        .stdout.strip()
        .splitlines()
    )

    assert (
        result_freetype == result_fontations
    ), f"FreeType and Fontations fc-query result must match. Fontations: {result_fontations}, FreeType: {result_freetype}"



#! /usr/bin/env python3
# Copyright (C) 2024 fontconfig Authors
# SPDX-License-Identifier: HPND

import os
import shutil
import subprocess
import sys
from pathlib import Path

sourcedir = os.environ.get('MESON_SOURCE_ROOT')
builddir = os.environ.get('MESON_BUILD_ROOT')
distdir = os.environ.get('MESON_DIST_ROOT')

if shutil.which('autoreconf'):
    print('no autoreconf installed', file=sys.stderr)

subprocess.run(['autoreconf', '-i'], cwd=distdir)

# Copy files for compatibility
for f in (Path(builddir) / 'doc').glob('*.1'):
    print(f'Copying {f.name}')
    shutil.copy2(f, Path(distdir) / f.stem)

# Remove autom4te.cache
shutil.rmtree(Path(distdir) / 'autom4te.cache')



#!/usr/bin/env python3

import os
import shutil

references = [
    'docs/Pango',
    'docs/PangoCairo',
    'docs/PangoFc',
    'docs/PangoFT2',
    'docs/PangoOT',
    'docs/PangoXft',
]

sourceroot = os.environ.get('MESON_SOURCE_ROOT')
buildroot = os.environ.get('MESON_BUILD_ROOT')
distroot = os.environ.get('MESON_DIST_ROOT')

for reference in references:
    src_path = os.path.join(buildroot, reference)
    if os.path.isdir(src_path):
        dst_path = os.path.join(distroot, reference)
        shutil.copytree(src_path, dst_path)



#! /usr/bin/env python3
# Copyright (C) 2024 fontconfig Authors
# SPDX-License-Identifier: HPND

import os
import shutil
from pathlib import Path

sourcedir = os.environ.get('MESON_SOURCE_ROOT')
builddir = os.environ.get('MESON_BUILD_ROOT')
distdir = os.environ.get('MESON_DIST_ROOT')

# Copy manpages
docdir = Path(distdir) / 'doc'
for f in (Path(builddir) / 'doc').glob('*.[135]'):
    print(f'Copying {f.name}')
    shutil.copy2(f, docdir)

# Copy config file
confdir = Path(distdir) / 'conf.d'
shutil.copy2(Path(builddir) / 'conf.d' / '35-lang-normalize.conf', confdir)

# Documentation
shutil.copy2(Path(builddir) / 'doc' / 'fontconfig-devel.html', docdir)
shutil.copy2(Path(builddir) / 'doc' / 'fontconfig-devel.pdf', docdir)
shutil.copy2(Path(builddir) / 'doc' / 'fontconfig-devel.txt', docdir)
shutil.copy2(Path(builddir) / 'doc' / 'fontconfig-user.html', docdir)
shutil.copy2(Path(builddir) / 'doc' / 'fontconfig-user.pdf', docdir)
shutil.copy2(Path(builddir) / 'doc' / 'fontconfig-user.txt', docdir)



#!/usr/bin/env python3

import os
import sys
import argparse
import platform
from pathlib import PurePath

if __name__=='__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('availpath')
    parser.add_argument('confpath')
    parser.add_argument('links', nargs='+')
    args = parser.parse_args()

    if os.path.isabs(args.confpath):
        destdir = os.environ.get('DESTDIR')
        if destdir:
            # c:\destdir + c:\prefix must produce c:\destdir\prefix
            confpath = str(PurePath(destdir, *PurePath(args.confpath).parts[1:]))
        else:
            confpath = args.confpath
    else:
        confpath = os.path.join(os.environ['MESON_INSTALL_DESTDIR_PREFIX'], args.confpath)

    if not os.path.exists(confpath):
        os.makedirs(confpath)

    for link in args.links:
        src = os.path.join(args.availpath, link)
        dst = os.path.join(confpath, link)
        try:
            os.remove(dst)
        except FileNotFoundError:
            pass
        try:
            os.symlink(os.path.relpath(src, start=args.confpath), dst)
        except NotImplementedError:
            # Not supported on this version of Windows
            break
        except OSError as e:
            # Symlink privileges are not available
            if platform.system().lower() == 'windows' and e.winerror == 1314:
                break
            raise



#!/usr/bin/env python3
#
# fontconfig write-35-lang-normalize-conf.py

import os
import sys

if len(sys.argv) < 2:
  print('ERROR: usage: {} ORTH_LIST [OUTPUT.CONF]'.format(sys.argv[0]))
  sys.exit(-1)

orth_list_unsorted = sys.argv[1].split(',')

if len(sys.argv) > 2 and sys.argv[2] != '-':
  f_out = open(sys.argv[2], 'w', encoding='utf8')
else:
  f_out = sys.stdout

orth_list = sorted(sys.argv[1].split(','))

print('<?xml version="1.0"?>', file=f_out)
print('<!DOCTYPE fontconfig SYSTEM "urn:fontconfig:fonts.dtd">', file=f_out)
print('<fontconfig>', file=f_out)

for o in orth_list:
  print(f'  <!-- {o}* -> {o} -->', file=f_out)
  print(f'  <match>', file=f_out)
  print(f'    <test name="lang" compare="contains"><string>{o}</string></test>', file=f_out)
  print(f'    <edit name="lang" mode="assign" binding="same"><string>{o}</string></edit>', file=f_out)
  print(f'  </match>', file=f_out)

print('</fontconfig>', file=f_out)

f_out.close()

sys.exit(0)



#!/usr/bin/env python3
#
# gperf compare_output.py
#
# Copyright (C) 2020 Tim-Philipp Müller <tim centricular com>
#
# This file is part of GNU GPERF.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Compares two text files line-by-line to make sure they match.

import os
import sys

if len(sys.argv) != 3:
  print('ERROR: usage: {} EXPECTED-OUTPUT-FILE ACTUAL-TEST-OUTPUT-FILE'.format(sys.argv[0]))
  sys.exit(-1)

fn_exp = sys.argv[1]
fn_test = sys.argv[2]
f_exp = open(fn_exp, 'r', encoding='utf8')
f_test = open(fn_test, 'r', encoding='utf8')

for (n, (line_exp, line_test)) in enumerate(zip(f_exp, f_test)):
  line_exp = line_exp.strip()
  line_test = line_test.strip()
  if line_exp != line_test:
    print('Output does not match in line {}:\nExpected: "{}" in {}\nActual:   "{}" in {}'.format(n+1, line_exp, fn_exp, line_test, fn_test))
    sys.exit(-1)

if f_exp.read() != '':
  print('Actual output file "{}" has fewer lines than expected out file "{}"!'.format(fn_test, fn_exp))
  sys.exit(-1)

if f_test.read() != '':
  print('Actual output file "{}" has more lines than expected out file "{}"!'.format(fn_test, fn_exp))
  sys.exit(-1)

f_exp.close()
f_test.close()

sys.exit(0)



#!/usr/bin/env python3

import argparse
import os
import shutil
import subprocess
import sys
import tempfile
from urllib.request import urlretrieve
import multiprocessing

import logging
import sys

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()

if not sys.stdout.isatty():
    logger.handlers = []

URL_TEMPLATE = (
    "https://commondatastorage.googleapis.com/chromeos-localmirror/distfiles/%s"
)
SOURCE_FILES = (
    "noto-cjk-20210501.tar.bz2",
    "crosextrafonts-20130214.tar.gz",
    "notofonts-20250214.tar.xz",
    "robotofonts-2.132.tar.bz2",
    "lohitfonts-cros-2.5.5.tar.bz2",
    "ko-nanumfonts-3.20.tar.bz2",
)
SOURCES = [URL_TEMPLATE % source for source in SOURCE_FILES]
SOURCES += ["https://github.com/googlefonts/roboto-flex/releases/download/3.100/roboto-flex-fonts.zip"]

CONTAINER_DOWNLOAD_DIR = "/testfonts"

# If you want to update these, copy the contents of the .stamp file from the target directory here.
EXPECTED_HASHES = """
noto-cjk-20210501.tar.bz2      5c70f246c392bf15aeaee22b2fba26bb0ac54e47c00a4b9c3b556bd06fe89281
crosextrafonts-20130214.tar.gz c48d1c2fd613c9c06c959c34da7b8388059e2408d2bb19845dc3ed35f76e4d09
notofonts-20250214.tar.xz      cb0f9edc030e07eca58b799c607d000dfde588911d8e228fc82d5f61b9e5b6b3
robotofonts-2.132.tar.bz2      087c6f0708fe71f71a056f70fdbd5a85f4d2ce916670a98bd4be10b168abe16a
lohitfonts-cros-2.5.5.tar.bz2  ce0ce2a5098c8ffc52327cc030576df7f5328ad9fd8a3289e2476990ad133ff1
ko-nanumfonts-3.20.tar.bz2     59f9b6d7fcf63ca2bea7156ad66c784a1f0601d47be1b11237e17733d7112832
roboto-flex-fonts.zip          02e0f5db84e69f254958434269d83aa6057b06f3c2a21042bb81b1afe1a0c8c6
"""

STAMP_FILE = ".stamp"


def download_file(url, tmp_dir, target_dir):
    filename = os.path.basename(url)
    filepath = os.path.join(tmp_dir, filename)
    logger.info(f"Downloading {url} to {filepath}")
    urlretrieve(url, filepath)
    return filepath


def compute_sha256(filepath):
    """Computes the SHA-256 hash of a file."""
    import hashlib

    hasher = hashlib.sha256()
    with open(filepath, "rb") as file:
        while True:
            chunk = file.read(4096)
            if not chunk:
                break
            hasher.update(chunk)
    return hasher.hexdigest()


def extract_archive(filepath, target_dir):
    """Extracts an archive to the target directory."""
    filename = os.path.basename(filepath)
    logger.info(f"Extracting {filepath} to {target_dir}")
    if filename.endswith((".tar.bz2", ".tar.gz", ".tar.xz")):
        subprocess.run(["tar", "xf", filepath, "-C", target_dir], check=True)
    elif filename.endswith(".zip"):
        shutil.unpack_archive(filepath, target_dir)
    else:
        raise ValueError(f"Unsupported archive type: {filename}")


def download_and_extract(url, tmp_dir, target_dir):
    archivepath = download_file(url, tmp_dir, target_dir)
    archivehash = compute_sha256(archivepath)
    extract_archive(archivepath, target_dir)
    return os.path.basename(archivepath), archivehash


def stamp_target_dir(target_dir, downloaded_files_hashes):
    if not downloaded_files_hashes or len(downloaded_files_hashes) == 0:
        logger.error("No files to stamp")
        return

    stamp_path = os.path.join(target_dir, STAMP_FILE)
    if os.path.exists(stamp_path):
        os.remove(stamp_path)

    max_filename_length = (
        max(len(filename) for filename, _ in downloaded_files_hashes)
        if downloaded_files_hashes
        else 0
    )

    hash_results = "\n".join(
        f"{filename:<{max_filename_length}} {hash}"
        for filename, hash in downloaded_files_hashes
    )
    with open(stamp_path, "w") as f:
        f.write(hash_results)

    if not stamp_hashes_match(stamp_path):
        raise ValueError("Downloaded files do not match recorded hashes!")


def stamp_hashes_match(stamp_path):
    if not os.path.exists(stamp_path):
        return False

    lines = []
    with open(stamp_path, "r") as f:
        lines = f.readlines()

    expected_hashes = {}
    for line in EXPECTED_HASHES.strip().split("\n"):
        filename, hash = line.split()
        expected_hashes[filename] = hash

    for line in lines:
        filename, hash = line.split()
        if filename not in expected_hashes:
            logger.error(f"Unexpected file in stamp: {filename}")
            return False
        if expected_hashes[filename] != hash:
            logger.error(
                f"Hash mismatch for {filename}: expected {expected_hashes[filename]}, got {hash}"
            )
            return False
    return True


def main():
    """Main function to parse arguments and download/extract fonts."""

    parser = argparse.ArgumentParser(description="Download and extract ChromeOS fonts.")

    parser.add_argument("--target-dir", help="Target directory for extracted fonts.")
    parser.add_argument(
        "--try-symlink",
        action="store_true",
        help="Try to symlink test files previously downloaded to the container.",
    )
    args = parser.parse_args()

    target_dir = args.target_dir

    if args.try_symlink:
        if os.path.exists(CONTAINER_DOWNLOAD_DIR) and stamp_hashes_match(
            os.path.join(CONTAINER_DOWNLOAD_DIR, STAMP_FILE)
        ):
            try:
                os.symlink(CONTAINER_DOWNLOAD_DIR, target_dir, target_is_directory=True)
            except FileExistsError:
                logger.debug(
                    f"Target directory {target_dir} already exists. Skipping symlink."
                )
            except OSError as e:
                logger.warning(f"Failed to create symlink: {e}")
            logger.info(
                f"Symlinked test fonts directory from {CONTAINER_DOWNLOAD_DIR} to {target_dir}."
            )
        else:
            logger.debug("Pre-downloaded test fonts not found.")

    if stamp_hashes_match(os.path.join(target_dir, STAMP_FILE)):
        logger.info("Fonts already downloaded and extracted.")
        return 0

    os.makedirs(target_dir, exist_ok=True)

    with tempfile.TemporaryDirectory() as tmp_dir:
        with multiprocessing.Pool(processes=multiprocessing.cpu_count()) as pool:
            hashes = pool.starmap(
                download_and_extract,
                [(source, tmp_dir, target_dir) for source in SOURCES],
            )
            stamp_target_dir(target_dir, hashes)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import argparse
import glob
import re
import sys
from pathlib import Path

ret = 0

parser = argparse.ArgumentParser()
parser.add_argument('path')

args = parser.parse_args()

for fn in glob.glob(str(Path(args.path) / '*.fncs')):
    with open(fn) as f:
        for i, line in enumerate(f):
            if re.search(r'[\w\[\]]+\s+@ARG', line, re.A):
                ret = 1
                print("{file}:{line}\t{str}".format(file=fn, line=i+1, str=line.rstrip()))

sys.exit(ret)



#!/usr/bin/env python3
#
# fontconfig/fc-lang/fc-lang.py
#
# Copyright © 2001-2002 Keith Packard
# Copyright © 2019 Tim-Philipp Müller
#
# Permission to use, copy, modify, distribute, and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice appear in all copies and that both that
# copyright notice and this permission notice appear in supporting
# documentation, and that the name of the author(s) not be used in
# advertising or publicity pertaining to distribution of the software without
# specific, written prior permission.  The authors make no
# representations about the suitability of this software for any purpose.  It
# is provided "as is" without express or implied warranty.
#
# THE AUTHOR(S) DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
# EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

# fc-lang
#
# Read a set of language orthographies and build C declarations for
# charsets which can then be used to identify which languages are
# supported by a given font.
#
# TODO: this code is not very pythonic, a lot of it is a 1:1 translation
# of the C code and we could probably simplify it a bit
import argparse
import string
import sys
import os

# we just store the leaves in a dict, we can order the leaves later if needed
class CharSet:
    def __init__(self):
        self.leaves = {} # leaf_number -> leaf data (= 16 uint32)

    def add_char(self, ucs4):
        assert ucs4 < 0x01000000
        leaf_num = ucs4 >> 8
        if leaf_num in self.leaves:
            leaf = self.leaves[leaf_num]
        else:
            leaf = [0, 0, 0, 0, 0, 0, 0, 0] # 256/32 = 8
            self.leaves[leaf_num] = leaf
        leaf[(ucs4 & 0xff) >> 5] |= (1 << (ucs4 & 0x1f))
        #print('{:08x} [{:04x}] --> {}'.format(ucs4, ucs4>>8, leaf))

    def del_char(self, ucs4):
        assert ucs4 < 0x01000000
        leaf_num = ucs4 >> 8
        if leaf_num in self.leaves:
            leaf = self.leaves[leaf_num]
            leaf[(ucs4 & 0xff) >> 5] &= ~(1 << (ucs4 & 0x1f))
            # We don't bother removing the leaf if it's empty */
            #print('{:08x} [{:04x}] --> {}'.format(ucs4, ucs4>>8, leaf))

    def equals(self, other_cs):
        keys = sorted(self.leaves.keys())
        other_keys = sorted(other_cs.leaves.keys())
        if len(keys) != len(other_keys):
            return False
        for k1, k2 in zip(keys, other_keys):
            if k1 != k2:
                return False
            if not leaves_equal(self.leaves[k1], other_cs.leaves[k2]):
                return False
        return True

# Convert a file name into a name suitable for C declarations
def get_name(file_name):
    return file_name.split('.')[0]

# Convert a C name into a language name
def get_lang(c_name):
    return c_name.replace('_', '-').replace(' ', '').lower()

def read_orth_file(file_name):
    lines = []
    with open(file_name, 'r', encoding='utf-8') as orth_file:
        for num, line in enumerate(orth_file):
            if line.startswith('include '):
                include_fn = line[8:].strip()
                lines += read_orth_file(include_fn)
            else:
                # remove comments and strip whitespaces
                line = line.split('#')[0].strip()
                line = line.split('\t')[0].strip()
                # skip empty lines
                if line:
                    lines += [(file_name, num, line)]

    return lines

def leaves_equal(leaf1, leaf2):
    for v1, v2 in zip(leaf1, leaf2):
        if v1 != v2:
            return False
    return True

# Build a single charset from a source file
#
# The file format is quite simple, either
# a single hex value or a pair separated with a dash
def parse_orth_file(file_name, lines):
    charset = CharSet()
    for fn, num, line in lines:
        delete_char = line.startswith('-')
        if delete_char:
            line = line[1:]
        if line.find('-') != -1:
            parts = line.split('-')
        elif line.find('..') != -1:
            parts = line.split('..')
        else:
            parts = [line]

        start = int(parts.pop(0), 16)
        end = start
        if parts:
            end = int(parts.pop(0), 16)
        if parts:
            print('ERROR: {} line {}: parse error (too many parts)'.format(fn, num))

        for ucs4 in range(start, end+1):
            if delete_char:
                charset.del_char(ucs4)
            else:
                charset.add_char(ucs4)

    assert charset.equals(charset) # sanity check for the equals function

    return charset

if __name__=='__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('orth_files', nargs='+', help='List of .orth files')
    parser.add_argument('--directory', dest='directory', default=None)
    parser.add_argument('--template', dest='template_file', default=None)
    parser.add_argument('--output', dest='output_file', default=None)

    args = parser.parse_args()

    sets = []
    names = []
    langs = []
    country = []

    total_leaves = 0

    LangCountrySets = {}

    # Open output file
    if args.output_file:
        sys.stdout = open(args.output_file, 'w', encoding='utf-8')

    # Read the template file
    if args.template_file:
        tmpl_file = open(args.template_file, 'r', encoding='utf-8')
    else:
        tmpl_file = sys.stdin

    # Change into source dir if specified (after opening other files)
    if args.directory:
        os.chdir(args.directory)

    orth_entries = {}
    for i, fn in enumerate(args.orth_files):
        orth_entries[fn] = i

    for fn in sorted(orth_entries.keys()):
        lines = read_orth_file(fn)
        charset = parse_orth_file(fn, lines)

        sets.append(charset)

        name = get_name(fn)
        names.append(name)

        lang = get_lang(name)
        langs.append(lang)
        if lang.find('-') != -1:
            country.append(orth_entries[fn]) # maps to original index
            language_family = lang.split('-')[0]
            if not language_family in LangCountrySets:
              LangCountrySets[language_family] = []
            LangCountrySets[language_family] += [orth_entries[fn]]

        total_leaves += len(charset.leaves)

    # Find unique leaves
    leaves = []
    for s in sets:
       for leaf_num in sorted(s.leaves.keys()):
           leaf = s.leaves[leaf_num]
           is_unique = True
           for existing_leaf in leaves:
               if leaves_equal(leaf, existing_leaf):
                  is_unique = False
                  break
           #print('unique: ', is_unique)
           if is_unique:
               leaves.append(leaf)

    # Find duplicate charsets
    duplicate = []
    for i, s in enumerate(sets):
        dup_num = None
        if i >= 1:
            for j, s_cmp in enumerate(sets):
                if j >= i:
                    break
                if s_cmp.equals(s):
                    dup_num = j
                    break

        duplicate.append(dup_num)

    tn = 0
    off = {}
    for i, s in enumerate(sets):
        if duplicate[i]:
            continue
        off[i] = tn
        tn += len(s.leaves)

    # Scan the input until the marker is found
    # FIXME: this is a bit silly really, might just as well hardcode
    #        the license header in the script and drop the template
    for line in tmpl_file:
        if line.strip() == '@@@':
            break
        print(line, end='')

    print('/* total size: {} unique leaves: {} */\n'.format(total_leaves, len(leaves)))

    print('#define LEAF0       ({} * sizeof (FcLangCharSet))'.format(len(sets)))
    print('#define OFF0        (LEAF0 + {} * sizeof (FcCharLeaf))'.format(len(leaves)))
    print('#define NUM0        (OFF0 + {} * sizeof (uintptr_t))'.format(tn))
    print('#define SET(n)      (n * sizeof (FcLangCharSet) + offsetof (FcLangCharSet, charset))')
    print('#define OFF(s,o)    (OFF0 + o * sizeof (uintptr_t) - SET(s))')
    print('#define NUM(s,n)    (NUM0 + n * sizeof (FcChar16) - SET(s))')
    print('#define LEAF(o,l)   (LEAF0 + l * sizeof (FcCharLeaf) - (OFF0 + o * sizeof (intptr_t)))')
    print('#define fcLangCharSets (fcLangData.langCharSets)')
    print('#define fcLangCharSetIndices (fcLangData.langIndices)')
    print('#define fcLangCharSetIndicesInv (fcLangData.langIndicesInv)')

    assert len(sets) < 65536 # FIXME: need to change index type to 32-bit below then

    print('''
static const struct {{
    FcLangCharSet  langCharSets[{}];
    FcCharLeaf     leaves[{}];
    uintptr_t      leaf_offsets[{}];
    FcChar16       numbers[{}];
    {}       langIndices[{}];
    {}       langIndicesInv[{}];
}} fcLangData = {{'''.format(len(sets), len(leaves), tn, tn,
                             'FcChar16 ', len(sets), 'FcChar16 ', len(sets)))

    # Dump sets
    print('{')
    for i, s in enumerate(sets):
        if duplicate[i]:
            j = duplicate[i]
        else:
            j = i
        print('    {{ "{}",  {{ FC_REF_CONSTANT, {}, OFF({},{}), NUM({},{}) }} }}, /* {} */'.format(
		langs[i], len(sets[j].leaves), i, off[j], i, off[j], i))

    print('},')

    # Dump leaves
    print('{')
    for l, leaf in enumerate(leaves):
        print('    {{ {{ /* {} */'.format(l), end='')
        for i in range(0, 8): # 256/32 = 8
            if i % 4 == 0:
                print('\n   ', end='')
            print(' 0x{:08x},'.format(leaf[i]), end='')
        print('\n    } },')
    print('},')

    # Dump leaves
    print('{')
    for i, s in enumerate(sets):
        if duplicate[i]:
            continue

        print('    /* {} */'.format(names[i]))

        for n, leaf_num in enumerate(sorted(s.leaves.keys())):
            leaf = s.leaves[leaf_num]
            if n % 4 == 0:
                print('   ', end='')
            found = [k for k, unique_leaf in enumerate(leaves) if leaves_equal(unique_leaf,leaf)] 
            assert found, "Couldn't find leaf in unique leaves list!"
            assert len(found) == 1
            print(' LEAF({:3},{:3}),'.format(off[i], found[0]), end='')
            if n % 4 == 3:
                print('')
        if len(s.leaves) % 4 != 0:
            print('')

    print('},')
	
    print('{')
    for i, s in enumerate(sets):
        if duplicate[i]:
            continue

        print('    /* {} */'.format(names[i]))

        for n, leaf_num in enumerate(sorted(s.leaves.keys())):
            leaf = s.leaves[leaf_num]
            if n % 8 == 0:
                print('   ', end='')
            print(' 0x{:04x},'.format(leaf_num), end='')
            if n % 8 == 7:
                print('')
        if len(s.leaves) % 8 != 0:
            print('')

    print('},')

    # langIndices
    print('{')
    for i, s in enumerate(sets):
        fn = '{}.orth'.format(names[i])
        print('    {}, /* {} */'.format(orth_entries[fn], names[i]))
    print('},')

    # langIndicesInv
    print('{')
    for i, k in enumerate(orth_entries.keys()):
        name = get_name(k)
        idx = names.index(name)
        print('    {}, /* {} */'.format(idx, name))
    print('}')

    print('};\n')

    print('#define NUM_LANG_CHAR_SET	{}'.format(len(sets)))
    num_lang_set_map = (len(sets) + 31) // 32;
    print('#define NUM_LANG_SET_MAP	{}'.format(num_lang_set_map))

    # Dump indices with country codes
    assert len(country) > 0
    assert len(LangCountrySets) > 0
    print('')
    print('static const FcChar32 fcLangCountrySets[][NUM_LANG_SET_MAP] = {')
    for k in sorted(LangCountrySets.keys()):
        langset_map = [0] * num_lang_set_map # initialise all zeros
        for entries_id in LangCountrySets[k]:
            langset_map[entries_id >> 5] |= (1 << (entries_id & 0x1f))
        print('    {', end='')
        for v in langset_map:
            print(' 0x{:08x},'.format(v), end='')
        print(' }}, /* {} */'.format(k))

    print('};\n')
    print('#define NUM_COUNTRY_SET {}\n'.format(len(LangCountrySets)))

    # Find ranges for each letter for faster searching
    # Dump sets start/finish for the fastpath
    print('static const FcLangCharSetRange  fcLangCharSetRanges[] = {\n')
    for c in string.ascii_lowercase: # a-z
        start = 9999
        stop = -1
        for i, s in enumerate(sets):
            if names[i].startswith(c):
                start = min(start,i)
                stop = max(stop,i)
        print('    {{ {}, {} }}, /* {} */'.format(start, stop, c))
    print('};\n')

    # And flush out the rest of the input file
    for line in tmpl_file:
        print(line, end='')
    
    sys.stdout.flush()



#!/usr/bin/env python3
#
# fontconfig/doc/run-quiet.py
#
# Runs command and discards anything it sends to stdout
#
# Copyright © 2020 Tim-Philipp Müller
#
# Permission to use, copy, modify, distribute, and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice appear in all copies and that both that
# copyright notice and this permission notice appear in supporting
# documentation, and that the name of the author(s) not be used in
# advertising or publicity pertaining to distribution of the software without
# specific, written prior permission.  The authors make no
# representations about the suitability of this software for any purpose.  It
# is provided "as is" without express or implied warranty.
#
# THE AUTHOR(S) DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
# EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
import subprocess
import sys
import os

if len(sys.argv) < 2:
  sys.exit('Usage: {} PROGRAM [ARGS..]'.format(sys.argv[0]))

command = sys.argv[1:]

with open(os.devnull, 'w') as out:
  sys.exit(subprocess.run(command, stdout=out).returncode)
